/* Copyright 2003  Alexander V. Diemand

    This file is part of MolTalk.

    MolTalk is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    MolTalk is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MolTalk; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */
 
/* vim: set filetype=objc: */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


#include "Atom.oh"
#include "Coordinates.oh"
#include "Matrix53.oh"

static NSDictionary *translateAtomnameToAtomtype=nil;


@implementation Atom


+(void)initialize	//@nodoc
{
    if (self == [Atom class])
    {
	if (!translateAtomnameToAtomtype)
	{
	    NSNumber *at_carbon = [NSNumber numberWithInt: ATOM_TYPE_CARBON];
	    NSNumber *at_nitrogen = [NSNumber numberWithInt: ATOM_TYPE_NITROGEN];
	    NSNumber *at_oxygen = [NSNumber numberWithInt: ATOM_TYPE_OXYGEN];
	    NSNumber *at_hydrogen = [NSNumber numberWithInt: ATOM_TYPE_HYDROGEN];
	    NSNumber *at_sulfur = [NSNumber numberWithInt: ATOM_TYPE_SULFUR];
	    NSNumber *at_phospho = [NSNumber numberWithInt: ATOM_TYPE_PHOSPHOROUS];
	    translateAtomnameToAtomtype = [NSDictionary dictionaryWithObjectsAndKeys:
							 at_carbon, @"CA",
							 at_carbon, @"C",
							 at_carbon, @"CB",
							 at_carbon, @"CG",
							 at_carbon, @"CG1",
							 at_carbon, @"CG2",
							 at_carbon, @"CD",
							 at_carbon, @"CD1",
							 at_carbon, @"CD2",
							 at_carbon, @"CE",
							 at_carbon, @"CE1",
							 at_carbon, @"CE2",
							 at_carbon, @"CE3",
							 at_carbon, @"CZ1",
							 at_carbon, @"CZ2",
							 at_carbon, @"CZ", 
							 at_carbon, @"CH2",
							 at_carbon, @"C2",
							 at_carbon, @"C4",
							 at_carbon, @"C5",
							 at_carbon, @"C5M",
							 at_carbon, @"C6",
							 at_carbon, @"C8",
							 at_carbon, @"C5*",
							 at_carbon, @"C4*",
							 at_carbon, @"C3*",
							 at_carbon, @"C2*",
							 at_carbon, @"C1*",
							 at_nitrogen, @"N",
							 at_nitrogen, @"ND1",
							 at_nitrogen, @"ND2",
							 at_nitrogen, @"NE1",
							 at_nitrogen, @"NE2",
							 at_nitrogen, @"NZ",
							 at_nitrogen, @"NE",
							 at_nitrogen, @"NH1",
							 at_nitrogen, @"NH2",
							 at_nitrogen, @"N1",
							 at_nitrogen, @"N2",
							 at_nitrogen, @"N3",
							 at_nitrogen, @"N4",
							 at_nitrogen, @"N6",
							 at_nitrogen, @"N7",
							 at_nitrogen, @"N9",
							 at_oxygen, @"O",
							 at_oxygen, @"OG",
							 at_oxygen, @"OG1",
							 at_oxygen, @"OE1",
							 at_oxygen, @"OE2",
							 at_oxygen, @"OD1",
							 at_oxygen, @"OD2",
							 at_oxygen, @"OH",
							 at_oxygen, @"OXT",
							 at_oxygen, @"O2",
							 at_oxygen, @"O4",
							 at_oxygen, @"O6",
							 at_oxygen, @"O1P",
							 at_oxygen, @"O2P",
							 at_oxygen, @"O5*",
							 at_oxygen, @"O4*",
							 at_hydrogen, @"1HD2",
							 at_hydrogen, @"2HD2",
							 at_phospho, @"P", 
							 at_sulfur, @"SG", 
							 at_sulfur, @"SD", 	nil];
	    RETAIN(translateAtomnameToAtomtype);
	}
    }
}


-(id)init	//@nodoc
{
    [super init];
    bonds=nil;
    name=nil;
    number=nil;
    return self;
}


-(void)dealloc	//@nodoc
{
	//NSLog(@"Atom__dealloc");
	if (bonds) {
		[bonds removeAllObjects];
		RELEASE(bonds);
	}
	if (name)
	{
    		RELEASE(name);
	}
	if (number)
	{
		RELEASE(number);
	}
	[super dealloc];
}


/*
 *   returns atom's serial number
 */
-(NSNumber*)number
{
	return number;
}


/*
 *   returns atom's name
 */
-(NSString*)name
{
	return name;
}


/*
 *   returns string of atom's number, name and coordinates
 */
-(NSString*)description
{
	return [NSString stringWithFormat:@"Atom:%@ %@ %@",number,name,[super description]];
}


/*
 *   derives the type of atom from its name
 */
-(int)type
{
	id tp = [translateAtomnameToAtomtype objectForKey: name];
	if (tp)
	{
		return [tp intValue];
	}
	return -1;
}


/*
 *   returns the atom's assigned temperature factor
 */
-(double)temperature
{
	return temperature;
}


/*
 *   add bond from atm2 to this atom
 */
-(void)bondTo:(Atom*)atm2
{
	if (!bonds)
	{
		bonds = [NSMutableDictionary new];
	}
	id bonded = [bonds objectForKey: [atm2 number]];
	if (bonded != nil)
	{ /* dont add twice */
		return;
	}
	//NSLog(@"adding bond from %@%@ to %@%@",[self number], [self name], [atm2 number], [atm2 name]);
	[bonds setObject: atm2 forKey: [atm2 number]];
}


/*
 *   returns enumerator over all bonded atoms
 */
-(NSEnumerator*)allBondedAtoms
{
	if (!bonds)
	{
		return nil;
	}
	return [[bonds allValues] reverseObjectEnumerator];
}


/*
 *   remove a bond from atm2 to this atom
 */
-(void)dropBondTo:(Atom*)atm2
{
	id bonded = [bonds objectForKey: [atm2 number]];
	if (bonded != nil)
	{
		[bonds removeObjectForKey: [atm2 number]];
	}
}


/*
 *   remove all bonds from this atom
 */
-(void)dropAllBonds
{
    if (bonds) {
	[bonds removeAllObjects];
	RELEASE(bonds);
    	bonds = nil;
    }
}


/*
 *   set serial number to the new one
 */
-(id)setNumber: (int)serial
{
	number = RETAIN([NSNumber numberWithInt: serial]);
	return self;
}


/*
 *   creation of an atom with number, name and explicitly setting coordinates
 */
+(Atom*)atomWithNumber:(unsigned int)num name:(char*)nm X:(double)x Y:(double)y Z:(double)z B:(double)b
{
	Atom *atom = [Atom new];
	[atom setX:x Y:y Z:z];
	atom->temperature = b;
	atom->number = RETAIN([NSNumber numberWithInt: num]);
	
	/* pad name */
	unsigned int len=strlen(nm);
	unsigned int i=0;
	while (nm[i]==' ')
	{
		i++;
		if (i>=(len-1))
		{
			break;
		}
	}
	unsigned int j=len-1;
	while (nm[j] == ' ')
	{
		j--;
		if (j==i)
		{
			break;
		}
	}
	nm[j+1]='\0';
	atom->name = RETAIN([NSString stringWithCString: (nm+i)]);
	
	//printf("Atom_atomWithNumber:%d name:%s X:%1.1f Y:%1.1f Z:%1.1f\n",num,nm,x,y,z);
	return AUTORELEASE(atom);
}


@end

